@@ -17,7 +17,8 @@
  * Boston, MA 02110-1301, USA.
  */
 
-#pragma once
+#ifndef __GST_D3D11_H264_DEC_H__
+#define __GST_D3D11_H264_DEC_H__
 
 #include "gstd3d11decoder.h"
 
@@ -25,7 +26,10 @@ G_BEGIN_DECLS
 
 void  gst_d3d11_h264_dec_register (GstPlugin * plugin,
                                    GstD3D11Device * device,
+                                   GstD3D11Decoder * decoder,
                                    guint rank,
                                    gboolean legacy);
 
 G_END_DECLS
+
+#endif /* __GST_D3D11_H264_DEC_H__ */
